"""
dict_product: a cartesian product of a dict.

Author: Erel Segal-Halevi
Since: 2022-05
"""

from typing import Dict, List, Any
import itertools

def dict_product(input:Dict[Any,List[Any]])->List[Dict[Any,Any]]:
    """
    Gets a dict mapping each key to a list of possible values.
    Generates all possible dicts, mapping each key to a single value.
    If, for each key k, there are nk possible values, then the number of dicts is prod_k(nk).
    >>> for x in dict_product({"x": [1,2,3], "y": [4,5], "z": [6]}): print(x)
    {'x': 1, 'y': 4, 'z': 6}
    {'x': 1, 'y': 5, 'z': 6}
    {'x': 2, 'y': 4, 'z': 6}
    {'x': 2, 'y': 5, 'z': 6}
    {'x': 3, 'y': 4, 'z': 6}
    {'x': 3, 'y': 5, 'z': 6}
    """
    keys  =  list(input.keys())
    value_lists = input.values()
    num_keys = len(keys)
    value_combinations = itertools.product(*value_lists)
    for combination in value_combinations:
        yield {keys[i]: combination[i] for i in range(num_keys)}


if __name__=="__main__":
    import doctest
    print(doctest.testmod())
    # for x in dict_product({"x": [1,2,3], "y": [4,5], "z": [6]}): print(x)




import pandas
from matplotlib import pyplot as plt
from experiments_csv.dict_to_row import dict_to_rows
# from typing import List

import logging
logger = logging.getLogger(__name__)

def plot_dataframe(ax, results: pandas.DataFrame,
    x_field:str, y_field:str, z_field:str, mean:bool=True, 
    legend_properties={'size':8}, xlim=None, ylim=None):
    """
    Plot on the given axis, results from the given pandas dataframe.

    :param ax      a matplotlib axis to plot on.
    :param results  a DataFrame containing the results to show.

    :param x_field: name of column for x axis.
    :param y_field: name of column for y axis.
    :param z_field: name of column for different lines in the same plot (for each value of this column, there will be a different line).
    :param mean: if True, it makes a line-plot of the mean over all rows with the same xcolumn and zcolumn. If False, it makes a scatter-plot of all values.

    :param legend_properties: delegated to the legend-drawing function.
    """
    if isinstance(z_field, list):
        z_field_join = ",".join(z_field)
        results[z_field_join] = results[z_field].apply(lambda x: ",".join(map(str,x)), axis=1)
        z_field=z_field_join
    
    z_values = results[z_field].unique()
    z_values.sort()
    for z_value in z_values:
        logger.info("    Line: %s=%s",z_field, z_value)
        results_for_z_value = results[results[z_field]==z_value]
        # label = z_value
        if isinstance(z_value,int):
            label = f"{z_field}={z_value}"
        else:
            label = z_value
        if mean:
            mean_results_for_z_value = results_for_z_value.groupby([x_field]).mean()
            ax.plot(mean_results_for_z_value.index, mean_results_for_z_value[y_field], label=label)
        else:
            ax.scatter(results_for_z_value[x_field], results_for_z_value[y_field], label=label)
        ax.legend(prop=legend_properties)

    if xlim is not None:
        try:
            ax.set_xlim(*xlim)
        except AttributeError:
            ax.xlim(*xlim)
    if ylim is not None:
        try:
            ax.set_ylim(*ylim)
        except AttributeError:
            ax.ylim(*ylim)


def single_plot_results(results_csv_file:str, filter:dict, 
    x_field:str, y_field:str, z_field:str, mean:bool=True, 
    save_to_file:bool=False,
    **kwargs):
    """
    Make a single plot of results from the given file.

    :param results_csv_file  path to the CSV file containing the results.
    :param filter: a dict used to filter the rows of the results file. See dict_to_row.

    :param x_field: name of column for x axis.
    :param y_field: name of column for y axis.
    :param z_field: name of column for different lines in the same plot (for each value of this column, there will be a different line).
    :param mean: if True, it makes a line-plot of the mean over all rows with the same xcolumn and zcolumn. If False, it makes a scatter-plot of all values.
    :param save_to_file: if True, it saves the plot to a PNG file with the same name as the results_csv_file. If False, it just shows the plot.

    :param kwargs: arguments to delegate to plot_dataframe.
    """
    results = pandas.read_csv(results_csv_file)
    results = dict_to_rows(results, filter)
    if len(filter)>0:
        title = ", ".join([f"{key}={value}" for key,value in filter.items()])
    else:
        title = "All results"
    logger.info("Plot: %s", title)
    plot_dataframe(plt, results, 
        x_field, y_field, z_field, mean, 
        **kwargs)

    plt.xlabel(x_field)
    ylabel = f"mean {y_field}" if mean else y_field
    plt.ylabel(ylabel)
    plt.title(title)

    if save_to_file:
        output_file = results_csv_file.replace(".csv",".png")
        plt.savefig(output_file)
    else:
        plt.show()




def multi_plot_results(results_csv_file:str, filter:dict,
    x_field:str, y_field:str, z_field:str, mean:bool, 
    subplot_field:str, subplot_rows:int, subplot_cols:int, 
    sharex: bool=False, sharey: bool=False,
    save_to_file:bool=False,
    **kwargs):
    """
    Make multiple plots (on subplots of the same figure), each time with a different value of the subplot_field column.

    :param results_csv_file  path to the CSV file containing the results.
    :param filter: a dict used to filter the rows of the results file. See dict_to_row.

    :param x_field: name of column for x axis.
    :param y_field: name of column for y axis.
    :param z_field: name of column for different lines in the same plot (for each value of this column, there will be a different line).
    :param mean: if True, it makes a line-plot of the mean over all rows with the same xcolumn and zcolumn. If False, it makes a scatter-plot of all values.

    :param subplot_field: name of column for different subplots (for each value of this column, there will be a different subplot).
    :param subplot_rows: num of rows in the subplots grid
    :param subplot_cols: num of cols in the subplots grid
    :param sharex: whether to share the x axis between subplots.
    :param sharey: whether to share the y axis between subplots.
    :param save_to_file: if True, it saves the plot to a PNG file with the same name as the results_csv_file. If False, it just shows the plot.

    :param kwargs: arguments to delegate to plot_dataframe.
    """
    results = pandas.read_csv(results_csv_file)
    results = dict_to_rows(results, filter)
    if len(filter)>0:
        suptitle = ", ".join([f"{key}={value}" for key,value in filter.items()])
    else:
        suptitle = "All results"
    logger.info("Plot: %s", suptitle)

    if isinstance(subplot_field, list):
        subplot_field_join = ",".join(subplot_field)
        results[subplot_field_join] = results[subplot_field].apply(lambda x: ",".join(map(str,x)), axis=1)
        subplot_field=subplot_field_join
    subplot_values = results[subplot_field].unique()
    subplot_values.sort()

    num_of_subpltos = subplot_rows * subplot_cols
    if num_of_subpltos < len(subplot_values):
        raise ValueError(f"Not enough subplots! {subplot_rows}*{subplot_cols} < len({subplot_values})")

    fig, axs = plt.subplots(subplot_rows, subplot_cols, sharex=sharex, sharey=sharey)
    axs = axs.flatten()
    axs_index = 0
    for subplot_value in subplot_values:
        subtitle = f"{subplot_field}={subplot_value}"
        logger.info("  Subplot: %s",subtitle)
        results_for_subplot_value = results[results[subplot_field]==subplot_value]
        plot_dataframe(axs[axs_index], results_for_subplot_value, 
            x_field, y_field, z_field, mean, 
            **kwargs)
        axs[axs_index].set_title(subtitle)
        axs_index += 1

    fig.supxlabel(x_field)
    ylabel = f"mean {y_field}" if mean else y_field
    fig.supylabel(ylabel)
    fig.suptitle(suptitle)

    if save_to_file:
        output_file = results_csv_file.replace(".csv",".png")
        fig.savefig(output_file)
    else:
        plt.show()


plot_dataframe.logger = single_plot_results.logger = multi_plot_results.logger = logger
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler())



"""
An abstract class that represents an experiment.
"""

from typing import Dict, List, Any, Callable
import pandas, os, logging, shutil, pathlib
from experiments_csv.dict_product import dict_product
from experiments_csv.dict_to_row import dict_to_row, dict_to_row_bounds
from datetime import datetime
from time import perf_counter
import inspect

import logging
logger = logging.getLogger(__name__)


SingleIterationFunction = Callable[[Any], Dict]
    # Represents a function that performs a single iteration of the experiment.
    # It accepts the run parameters (independent variables), and returns a dict with the run outcomes.


class Experiment:

    def __init__(self, results_folder="results", results_filename="results.csv", backup_folder="results_backup"):
        """
        :param columns: an ordered list of column names for this table.
        :param results_file: a path to a file that will be used to backup this table.

        If the file exists, it is read into the new table.
        If the file does not exist, an empty table is initialized and the file will be created when the table is "done".
        """
        pathlib.Path(results_folder).mkdir(parents=True, exist_ok=True)            
        results_file = os.path.join(results_folder, results_filename)
        if os.path.isfile(results_file):
            # Load existing file:
            self.dataFrame = pandas.read_csv(results_file)
            logger.info("Loaded %d rows from %s.", self.dataFrame.shape[0], results_file)

            # Backup existing file into the 'backup' folder:
            if backup_folder is not None:
                pathlib.Path(backup_folder).mkdir(parents=True, exist_ok=True)
                modification_timestamp = os.path.getmtime(results_file)
                modification_datetime = datetime.fromtimestamp(modification_timestamp).strftime("%Y_%m_%d__%H_%M_%S")
                backup_filename = results_filename.replace(".csv",f".{modification_datetime}.csv")
                backup_file = os.path.join(backup_folder, backup_filename)
                shutil.copyfile(results_file, backup_file)

        else:
            self.dataFrame = None  # will be initialized when the first row is added
            logger.info("Initialized an empty DataFrame bound to %s.", results_file)
        self.results_file = results_file

    def clear_previous_results(self):
        self.dataFrame = None

    def to_csv(self, filename):
        try:
            self.dataFrame.to_csv(filename, columns=self.dataFrame.columns, index=False)
            # index=False means to not save the automatically-added index column.
            # This is important - without it we might have problems later when adding new rows after load.
        except OSError as err:
            logger.warn(f"OS error {err}")

    def add(self, new_row:dict):
        """
        Add a data-row whose values exactly match the columns in self.columns.
        """
        if self.dataFrame is None:
            columns = list(new_row.keys())
            logger.debug("    Inserting first row: setting columns to %s", columns)
            self.dataFrame = pandas.DataFrame(columns=columns)

        index_of_new_row = self.dataFrame.shape[0]
        self.dataFrame.loc[index_of_new_row] = pandas.Series(new_row) # Works only if there are no "holes" in the index (requires "index=False" in saving to csv).
        # self.dataFrame.iloc[index_of_new_row] = pandas.Series(dataRow) # IndexError: single positional indexer is out-of-bounds
        # self.dataFrame = self.dataFrame.append(pandas.Series(dataRow), ignore_index=True) # This is inefficient! It creates and returns a new dataFrame.
        self.to_csv(self.results_file)


    def run(self, 
        single_run: SingleIterationFunction,
        input_ranges: Dict[str,List[Any]],
        ):
        """
        Runs the experiment, changing each parameter in the given range.

        :param single_run:   a function that performs a single run. It accepts the run parameters (independent variables), and returns a dict with the run outcomes.
        :param input_ranges: a dict where the key is the parameter name, and the value is a list of possible values for that parameter.
        """
        for input in dict_product(input_ranges):
            input_normalized = {k:normalized(v) for k,v in input.items()}
            logger.info("\nInput: %s", input_normalized)
            if self.dataFrame is not None:  # Check if there is an existing row:
                try:
                    existing_row = dict_to_row(self.dataFrame, input_normalized)
                except KeyError as err:
                    raise KeyError(f"You sent an input field that does not have a column in the existing CSV file. Please start a new CSV file. Error: ", {err} )
                if existing_row:
                    logger.info("Skipped existing row: %s", existing_row)
                    continue
            output = single_run(**input)
            if not isinstance(output, dict):
                raise ValueError(f"single_run must return a dict output, mapping each output variable name to its value. It returned {type(output)}.")
            logger.info("Output: %s", output)
            self.add({**input_normalized, **output})
        logger.info("\nDone!")


    def run_with_time_limit(self, 
        single_run: SingleIterationFunction,
        input_ranges: Dict[str,List[Any]],
        time_limit: float,
        runtime_field_name="runtime"
        ):
        """
        Runs the experiment, changing each parameter in the given range.
        Measures the run-time of each single run. If the run-time is above the given time-limit,
            it will not run experiments with larger input values.

        :param single_run:   a function that performs a single run. It accepts the run parameters (independent variables), and returns a dict with the run outcomes.
        :param input_ranges: a dict where the key is the parameter name, and the value is a list of possible values for that parameter.
        :param time_limit: the maximum time for a single run.
        :param runtime_field_name: the name of the column for storing the run-time.
        """

        for input in dict_product(input_ranges):
            input_normalized = {k:normalized(v) for k,v in input.items()}
            logger.info("\nInput: %s", input_normalized)
            if self.dataFrame is not None:  # Check if there is an existing row:
                try:
                    existing_row = dict_to_row(self.dataFrame, input_normalized)
                except KeyError as err:
                    raise KeyError(f"You sent an input field that does not have a column in the existing CSV file. Please start a new CSV file. Error: ", {err} )
                if existing_row:
                    logger.info("Existing row skipped: %s", existing_row)
                    continue
                dominating_row = dict_to_row_bounds(self.dataFrame, lowerbound={runtime_field_name: time_limit}, upperbound=input_normalized)
                if dominating_row:
                    logger.info("Skipped a combination that would probably take longer than the time-limit %f. Existing row: %s", time_limit, dominating_row)
                    continue
            
            time_before = perf_counter()
            output = single_run(**input)
            runtime = perf_counter() - time_before
            output[runtime_field_name] = runtime

            if not isinstance(output, dict):
                raise ValueError(f"single_run must return a dict output, mapping each output variable name to its value. It returned {type(output)}.")
            logger.info("Output: %s", output)
            self.add({**input_normalized, **output})
        logger.info("\nDone!")



def normalized(value):
    """
    Return a normalized version of the given value, for insertion into the table.
    """
    if hasattr(value,'__name__'):
        return value.__name__
    elif hasattr(value,'__dict__'):  # value is an object of some class - use the class "__str__" method
        return str(value)
    else:
        return value


Experiment.logger = logger
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler())



import pathlib
HERE = pathlib.Path(__file__).parent
__version__ = (HERE / "VERSION").read_text().strip()

from experiments_csv.Experiment import Experiment, logger
from experiments_csv.plot_results import plot_dataframe, single_plot_results, multi_plot_results



import pathlib
import setuptools

NAME = "experiments_csv"
URL = "https://github.com/erelsgl/" + NAME
HERE = pathlib.Path(__file__).parent
print(f"\nHERE = {HERE.absolute()}\n")
README = (HERE / "README.md").read_text()
REQUIRES = (HERE / "requirements.txt").read_text().strip().split("\n")
REQUIRES = [lin.strip() for lin in REQUIRES]
print(f'\nVERSION = {(HERE / NAME / "VERSION").absolute()}\n')
VERSION = (HERE / NAME / "VERSION").read_text().strip()
# See https://packaging.python.org/en/latest/guides/single-sourcing-package-version/

setuptools.setup(
    name=NAME,
    packages=setuptools.find_packages(),
    version=VERSION,
    install_requires=REQUIRES,
    extras_require = {
        "plotting": ["matplotlib"]
    },
    author="Erel Segal-Halevi",
    author_email="erelsgl@gmail.com",
    description="Simple framework for running simulation experiments and recording them in a CSV file",
    keywords="experiments",
    license="MIT",
    license_files=("LICENSE",),
    long_description=README,
    long_description_content_type="text/markdown",
    url=URL,
    project_urls={"Bug Reports": URL + "/issues", "Source Code": URL},
    python_requires=">=3.8",
    include_package_data=True,
    classifiers=[
        # see https://pypi.org/classifiers/
        "Development Status :: 2 - Pre-Alpha",
    ],
)

# Build:
#   Delete old folders: build, dist, *.egg_info
#   Then run:
#        python -m build
#   Or (old version):
#        python setup.py sdist bdist_wheel

# Publish to test PyPI:
#   twine upload --repository testpypi dist/*

# Publish to real PyPI:
#   twine upload --repository pypi dist/*



"""
dict_key: allows you to use a dict as a key for selecting a row in a pandas dataframe.

Author: Erel Segal-Halevi
Since: 2022-05
"""

import pandas

from numbers import Number

def dict_to_rows(df: pandas.DataFrame, key:dict)->pandas.DataFrame:
    """
    Accepts a DataFrame and a dict mapping column-names to values.
    Searches for rows in which the given columns contain the given values.
    Returns a dataframe with only these rows.

    >>> df = pandas.DataFrame({'a': [1,4,1], 'b': [2,5,5], 'c':[3,6,9], 'z':[123, 456, 159]})
    >>> dict_to_rows(df, {"a":1})
       a  b  c    z
    0  1  2  3  123
    2  1  5  9  159
    >>> dict_to_rows(df, {"a":1, "b":2, "c":3})
       a  b  c    z
    0  1  2  3  123
    >>> dict_to_rows(df, {"a":1, "b":2, "c":9})
    Empty DataFrame
    Columns: [a, b, c, z]
    Index: []
    """
    for k,v in key.items():
        df = df[df[k]==v]
    return df

def dict_to_row(df: pandas.DataFrame, key:dict)->dict:
    """
    Accepts a DataFrame and a dict mapping column-names to values.
    Searches a row in which the given columns contain the given values.
    Returns one such row if it exists (as a dict), or None if no such row exists.

    >>> df = pandas.DataFrame({'a': [1,4,1], 'b': [2,5,5], 'c':[3,6,9], 'z':[123, 456, 159]})
    >>> dict_to_row(df, {"a":1, "b":2, "c":3})
    {'a': 1, 'b': 2, 'c': 3, 'z': 123}
    
    >>> dict_to_row(df, {"a":1, "b":2, "c":9}) is None
    True
    """
    df = dict_to_rows(df, key)
    if df.empty:
        return None
    else:
        return df.iloc[0].to_dict()

def dict_to_rows_bounds(df: pandas.DataFrame, lowerbound: dict = {}, upperbound: dict = {}):
    """
    Accepts a DataFrame and two dict mapping column-names to values.
    Searches a row in which:
    * Each of the given *numeric* columns contains *at least* the values given in lowerbound, and *at most* the values given in upperbound.
    * Each of the given *non-numeric* columns contains *exactly* the given values.
    Returns a dataframe with only these rows.

    >>> df = pandas.DataFrame({'a': [1,4,1], 'b': [2,5,5], 'c':['tt','uu','vv'], 'z':[123, 456, 159], 't':[True, False, True]})
    >>> dict_to_rows_bounds(df, upperbound={"a":1, "b":2, "c":'tt'})
       a  b   c    z     t
    0  1  2  tt  123  True
    >>> dict_to_rows_bounds(df, upperbound={"a":1, "b":3, "c":'tt'})
       a  b   c    z     t
    0  1  2  tt  123  True
    >>> dict_to_rows_bounds(df, upperbound={"a":5, "b":6, "c":'uu'})
       a  b   c    z      t
    1  4  5  uu  456  False
    >>> dict_to_rows_bounds(df, upperbound={"a":2, "b":1, "c":'tt'}).empty
    True
    >>> dict_to_rows_bounds(df, upperbound={"a":1, "b":2, "c":'xx'}).empty
    True
    >>> dict_to_rows_bounds(df, lowerbound={"a":2, "b":1, "c":'uu'})
       a  b   c    z      t
    1  4  5  uu  456  False
    >>> dict_to_rows_bounds(df, lowerbound={"t":True})
       a  b   c    z     t
    0  1  2  tt  123  True
    2  1  5  vv  159  True
    >>> dict_to_rows_bounds(df, lowerbound={"t":False})
       a  b   c    z      t
    1  4  5  uu  456  False
    """
    for k,v in upperbound.items():
        if isnumber(v):
            df = df[df[k]<=v]
        else:
            df = df[df[k]==v]
    for k,v in lowerbound.items():
        if isnumber(v):
            df = df[df[k]>=v]
        else:
            df = df[df[k]==v]
    return df


def dict_to_row_bounds(df: pandas.DataFrame, lowerbound: dict = {}, upperbound: dict = {}):
    """
    Accepts a DataFrame and two dict mapping column-names to values.
    Searches a row in which:
    * Each of the given *numeric* columns contains *at least* the values given in lowerbound, and *at most* the values given in upperbound.
    * Each of the given *non-numeric* columns contains *exactly* the given values.

    Returns one such row if it exists, or None if no such row exists.

    >>> df = pandas.DataFrame({'a': [1,4,1], 'b': [2,5,5], 'c':['tt','uu','vv'], 'z':[123, 456, 159], 't':[True, False, True]})
    >>> dict_to_row_bounds(df, upperbound={"a":1, "b":2, "c":'tt'})
    {'a': 1, 'b': 2, 'c': 'tt', 'z': 123, 't': True}
    >>> dict_to_row_bounds(df, upperbound={"a":1, "b":3, "c":'tt'})
    {'a': 1, 'b': 2, 'c': 'tt', 'z': 123, 't': True}
    >>> dict_to_row_bounds(df, upperbound={"a":5, "b":6, "c":'uu'})
    {'a': 4, 'b': 5, 'c': 'uu', 'z': 456, 't': False}
    >>> dict_to_row_bounds(df, upperbound={"a":2, "b":1, "c":'tt'}) is None
    True
    >>> dict_to_row_bounds(df, upperbound={"a":1, "b":2, "c":'xx'}) is None
    True
    >>> dict_to_row_bounds(df, lowerbound={"a":2, "b":1, "c":'uu'})
    {'a': 4, 'b': 5, 'c': 'uu', 'z': 456, 't': False}
    """
    df = dict_to_rows_bounds(df, lowerbound, upperbound)
    if df.empty:
        return None
    else:
        return df.iloc[0].to_dict()


def isnumber(v)->bool:
    return isinstance(v, Number) and not isinstance(v, bool)

if __name__=="__main__":
    import doctest
    print(doctest.testmod())

